# Copyright (C) 2016, 2017 Cumulus Networks, Inc. All rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt
from oslo.config import cfg
from restapi.apps.ml2.discovery import DiscoveryManager
from restapi.utils.execute import execute_command
from restapi.utils.execute import execute_command_no_resp
from subprocess import check_output
import os
import uuid


INVALID_PID = 0

# TODO - What's the point of having a class with nothing but class methods?
# Get rid of the class.


class L2Manager(object):
    __BRIDGE_INTERFACES_FS = '/sys/devices/virtual/net/{bridge}/brif/'

    __BRIDGE_NAME_PREFIX = 'br'

    __VXLAN_NAME_PREFIX = 'vxlan'

    __SUBINTERFACE_NAME = '{interface}.{vlan}'

    __INTERFACE_BOND_SLAVE = '/sys/class/net/{intf}/bonding_slave'

    @classmethod
    def __net_commit(cls, newbridge, resp):
            cmd = 'net commit'
            execute_command(cmd, resp)

    @classmethod
    def __device_exists(cls, device_name):
        cmd = 'ip link show dev {device}'.format(device=device_name)
        result, status = execute_command_no_resp(cmd)
        if result is None:
            return False
        return status

    @classmethod
    def __interface_exists_on_bridge(cls, bridge_name, interface):
        directory = cls.__BRIDGE_INTERFACES_FS.format(bridge=bridge_name)
        if not os.path.isdir(directory):
            return False

        for filename in os.listdir(directory):
            if filename == interface:
                return True
        return False

    @classmethod
    def __up_device(cls, device_name):
        cmd = 'ip link set {device} up'.format(device=device_name)
        execute_command_no_resp(cmd)

    @classmethod
    def __down_device(cls, device_name):
        cmd = 'ip link set {device} down'.format(device=device_name)
        execute_command_no_resp(cmd)

    @classmethod
    def __delete_device(cls, device_name, resp, is_bridge):
        if not cls.__device_exists(device_name):
            return

        cls.__down_device(device_name)
        cmd = 'ip link del {device}'.format(device=device_name)
        execute_command(cmd, resp)

    @classmethod
    def get_bridge_name(cls, network_id):
        bridge_name = cls.__BRIDGE_NAME_PREFIX + network_id[:12]
        return bridge_name

    @classmethod
    def get_subinterface_name(cls, physical_interface, vlan_id):
        return cls.__SUBINTERFACE_NAME.format(interface=physical_interface,
                                              vlan=vlan_id)

    @classmethod
    def get_vxlan_name(cls, vni):
        return cls.__VXLAN_NAME_PREFIX + str(vni)

    @classmethod
    def convert_host_to_interface(cls, host):
        interface_name = DiscoveryManager.find_interface(host)
        directory = cls.__INTERFACE_BOND_SLAVE.format(intf=interface_name)
        if os.path.isdir(directory):
            cmd = '/sys/class/net/{intf}/master'.format(intf=interface_name) 
            result = os.readlink(cmd)
            if result:
                result = result.rstrip()
                interface_name = result.split("virtual/net/", 1)[1]

        return interface_name

    @classmethod
    def add_vlan(cls, physical_interface, vlan_id, resp):
        interface = cls.get_subinterface_name(physical_interface, vlan_id)
        if not cls.__device_exists(interface):
            cmd = ('ip link add link {phys} name {device} type vlan id'
                   ' {vid}'.format(phys=physical_interface,
                                   device=interface,
                                   vid=vlan_id))
            execute_command(cmd, resp)
            cls.__up_device(interface)

        return interface

    @classmethod
    def delete_vlan(cls, interface_name, resp):
        cls.__delete_device(interface_name, resp, False)

    @classmethod
    def add_interface(cls, bridge_name, interface_name, vlan_id,
                      newbridge, vxlan, resp):
        exists = False
        if cls.__interface_exists_on_bridge(bridge_name, interface_name):
            if newbridge:
                exists = True
            else:
                return

        if not exists:
            if cfg.CONF.ML2.use_net and newbridge:
                cmd = ('net add bridge {bridge} ports {interface}'.
                       format(bridge=bridge_name, interface=interface_name))
            else:
                cmd = ('brctl addif {bridge} {interface}'.
                       format(bridge=bridge_name, interface=interface_name))
            execute_command(cmd, resp)

        if newbridge:
            if vxlan:
                if cfg.CONF.ML2.use_net:
                    cmd = ('net add vxlan {interface} bridge access {vlan}'.
                           format(interface=interface_name, vlan=vlan_id))
                else:
                    cmd = ('bridge vlan del vid 1 untagged pvid dev '
                           '{interface}'.format(interface=interface_name))
                    execute_command(cmd, resp)
                    cmd = ('bridge vlan add vid {vlan} untagged pvid dev '
                           '{interface}'.format(vlan=vlan_id,
                                                interface=interface_name))
            else:
                if cfg.CONF.ML2.use_net:
                    cmd = ('net add interface {interface} bridge vids {vlan}'.
                           format(interface=interface_name, vlan=vlan_id))
                else:
                    cmd = ('bridge vlan add vid {vlan} dev {interface}'.
                           format(vlan=vlan_id, interface=interface_name))
            execute_command(cmd, resp)

            if cfg.CONF.ML2.use_net:
                cls.__net_commit(newbridge, resp)

    @classmethod
    def delete_interface(cls, bridge_name, physical_interface, vlan_id,
                         newbridge, is_delete, resp):
        if newbridge:
            interface_name = physical_interface
        else:
            interface_name = cls.get_subinterface_name(physical_interface,
                                                       vlan_id)

        if not cls.__interface_exists_on_bridge(bridge_name, interface_name):
            return interface_name

        if newbridge:
            if cfg.CONF.ML2.use_net:
                cmd = ('net del interface {interface} bridge vids {vlan}'.
                       format(interface=interface_name, vlan=vlan_id))
            else:
                cmd = ('bridge vlan del vid {vlan} dev {interface}'.
                       format(vlan=vlan_id, interface=interface_name))
            execute_command(cmd, resp)

        if is_delete:
            if cfg.CONF.ML2.use_net and newbridge:
                cmd = ('net del bridge {bridge} ports {interface}'.
                       format(bridge=bridge_name, interface=interface_name))
            else:
                cmd = ('brctl delif {bridge} {interface}'.
                       format(bridge=bridge_name, interface=interface_name))
            execute_command(cmd, resp)

        if newbridge and cfg.CONF.ML2.use_net:
            cls.__net_commit(newbridge, resp)

        return interface_name

    @classmethod
    def add_bridge(cls, bridge_name, vlan_id, spf, newbridge, resp):
        bridge_exists = False

        if cls.__device_exists(bridge_name):
            if newbridge:
                bridge_exists = True
            else:
                return

        if newbridge:
            if not cfg.CONF.ML2.use_net:
                if not bridge_exists:
                    cmd = 'brctl addbr {bridge}'.format(bridge=bridge_name)
                    execute_command(cmd, resp)

                    cmd = 'ip link set {bridge} type bridge vlan_filtering 1'\
                          .format(bridge=bridge_name)
                    execute_command(cmd, resp)

                    if spf:
                        cmd = 'brctl stp {bridge} on'\
                              .format(bridge=bridge_name)
                    else:
                        cmd = 'brctl stp {bridge} off'\
                              .format(bridge=bridge_name)
                    execute_command(cmd, resp)

                cmd = 'bridge vlan add vid {vlan} dev {bridge} self'\
                      .format(vlan=vlan_id, bridge=bridge_name)
                execute_command(cmd, resp)

            else:
                if not bridge_exists:
                    cmd = 'net add bridge'
                    execute_command(cmd, resp)

                if not spf:
                    cmd = 'net add bridge stp off'
                    execute_command(cmd, resp)

                cmd = 'net add bridge {bridge} vids {vlan}'\
                      .format(bridge=bridge_name, vlan=vlan_id)
                execute_command(cmd, resp)

                cls.__net_commit(newbridge, resp)
        else:
            cmd = 'brctl addbr {bridge}'.format(bridge=bridge_name)
            execute_command(cmd, resp)
            if spf:
                cmd = 'brctl stp {bridge} on'.format(bridge=bridge_name)
            else:
                cmd = 'brctl stp {bridge} off'.format(bridge=bridge_name)
            execute_command(cmd, resp)

        if cfg.CONF.ML2.trunk_interfaces:
            for trunk in cfg.CONF.ML2.trunk_interfaces:
                if newbridge:
                    trunk_interface_name = trunk
                else:
                    trunk_interface_name = cls.add_vlan(trunk, vlan_id, resp)

                cls.add_interface(bridge_name, trunk_interface_name,
                                  vlan_id, newbridge, False, resp)
        cls.__up_device(bridge_name)

    @classmethod
    def delete_bridge(cls, bridge_name, vlan_id, newbridge, refcnt, resp):
        if newbridge:
            if cfg.CONF.ML2.use_net:
                cmd = 'net del bridge {bridge} vids {vlan}'\
                      .format(bridge=bridge_name, vlan=vlan_id)
            else:
                cmd = 'bridge vlan del vid {vlan} dev {bridge} self'\
                      .format(vlan=vlan_id, bridge=bridge_name)
            execute_command(cmd, resp)

        if refcnt == 1:
            if not newbridge:
                cls.__delete_device(bridge_name, resp, True)
            for trunk in cfg.CONF.ML2.trunk_interfaces:
                if newbridge:
                    cls.delete_interface(bridge_name, trunk, vlan_id,
                                         newbridge, True, resp)
                else:
                    interface_name = cls.get_subinterface_name(trunk, vlan_id)
                    cls.delete_vlan(interface_name, resp)
        else:
            if newbridge:
                for trunk in cfg.CONF.ML2.trunk_interfaces:
                    cls.delete_interface(bridge_name, trunk, vlan_id,
                                         newbridge, False, resp)

    @classmethod
    def add_vxlan(cls, vni, bridge_name, vlan_id, newbridge, resp):
        vxlan_name = cls.get_vxlan_name(vni)

        if cls.__device_exists(vxlan_name):
            return

        if newbridge and cfg.CONF.ML2.use_net:
            cmd = 'net add vxlan {vxlan} vxlan id {vni}'\
                  .format(vxlan=vxlan_name, vni=vni)
            execute_command(cmd, resp)

            cmd = 'net add vxlan {vxlan} vxlan local-tunnelip {bind_ip}'\
                  .format(vxlan=vxlan_name, bind_ip=cfg.CONF.ML2.local_bind)
            execute_command(cmd, resp)

            cmd = 'net add vxlan {vxlan} vxlan remoteip {rem_ip}'\
                  .format(vxlan=vxlan_name, rem_ip=cfg.CONF.ML2.service_node)
            execute_command(cmd, resp)
        else:
            cmd = ('ip link add {vxlan} type vxlan id {vni} local {bind_ip} '
                   'remote {rem_ip}'.format(vxlan=vxlan_name,
                                            vni=vni,
                                            bind_ip=cfg.CONF.ML2.local_bind,
                                            rem_ip=cfg.CONF.ML2.service_node))
            execute_command(cmd, resp)

        cls.add_interface(bridge_name, vxlan_name,
                          vlan_id, newbridge, True, resp)

        cls.__up_device(vxlan_name)

    @classmethod
    def delete_vxlan(cls, vni, newbridge, resp):
        vxlan_name = cls.get_vxlan_name(vni)
        if not cls.__device_exists(vxlan_name):
            return

        cls.__down_device(vxlan_name)

        if newbridge and cfg.CONF.ML2.use_net:
            cmd = 'net del vxlan {vxlan}'.format(vxlan=vxlan_name)
            execute_command(cmd, resp)

            cls.__net_commit(newbridge, resp)
        else:
            cmd = 'ip link del {vxlan}'.format(vxlan=vxlan_name)
            execute_command(cmd, resp)


class HashManager(object):

    def __init__(self):
        self.pid = os.getpid()
        self.net_pid = self.get_networking_pid()
        self.hashid = uuid.uuid1()

    def get_networking_pid(self):
        # TODO - Is "shell=True" necessary?
        output = check_output("systemctl status networking.service",
                              shell=True)
        for row in output.splitlines():
            if 'Main PID' in row:
                key, value = row.split(': ')
                pid, info = value.split(' (')
                return int(pid)

        return INVALID_PID

    def get_hash_id(self):
        curr_net_pid = self.get_networking_pid()
        if curr_net_pid != self.net_pid:
            self.hashid = uuid.uuid1()
        return str(self.hashid)
