# Copyright (C) 2017 Cumulus Networks, Inc. All rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from base64 import b64decode
from subprocess import CalledProcessError, check_output, STDOUT
import falcon
import json
import os
import pwd


def make_switch_user_closure(user):
    """
    Return a function that sets the process user and group IDs to those of a
    given user.
    """

    def switch_user():
        """Set the process user and group ID to that of an enclosed user."""

        _, _, uid, gid, _, _, _ = pwd.getpwnam(user)

        # Change the gid before changing the uid to avoid raising OSError!
        os.setgid(gid)
        os.setuid(uid)

    return switch_user


class Rpc(object):
    """
    This resource is the NCLU client, net(.py).  Post a command to this resource
    to get the same effect as running that command in a shell.
    """

    def on_post(self, req, resp):
        """
        HTTP POST requests must have the "Content-Type" header set to "application/json".  Failure
        to set this header will result in Falcon returning an HTTP status 400 response.

        For testing from the command line with self-signed server certificates and HTTP basic authentication:
            http --verify=no -j --auth <username>:<password> POST https://<server ip>:8080/nclu/v1/rpc cmd='show counters'
            curl -D - -X POST -H "Content-Type: application/json" -k -u <username>:<password> -d '{"cmd": "show counters"}'  https://<server ip>:8080/nclu/v1/rpc
        """

        if req.content_length == 0:
            raise falcon.HTTPBadRequest('No POST data.',
                "You must POST a 'net' command as application/json: '{\"cmd\": \"<net command>\"}'.  Set the Content-Type header to application/json.")

        try:
            # bytes --> string --> json
            post_data = json.loads(req.stream.read(req.content_length).decode())
#        except json.JSONDecodeError:    <-- new in Python 3.5; we use 3.4 in CL currently
        except ValueError:  # TODO - Delete this line when the above is supported.
            raise falcon.HTTPBadRequest('Invalid POST data.',
                "You must POST a valid 'net' command as ascii text with the Content-Type header set to application/json.")

        assert isinstance(post_data, dict)
        if 'cmd' not in post_data:
            raise falcon.HTTPBadRequest('Invalid POST data.', "You must POST a 'net' command as the value of the key 'cmd'.")
        if post_data['cmd'].lower().startswith('net'):
            raise falcon.HTTPBadRequest("Don't include 'net'.", "Don't include 'net' with the command string.")

        # If the request includes an Authorization header, run NCLU as the given user.
        # Note that credentials, if present, have already be verified upstream via Nginx configuration.
        # If credentials were not required, Nginx shoud have removed the Authorization header.
        pre_exec = None
        if req.auth is not None:
            try:
                # req.auth looks like "Basic Yadslkfj49348" (without the quotes).
                b64 = req.auth.split()[1]

                # The decoded b64 looks like "user:pass" (without the quotes).
                user = b64decode(b64).split(':')[0]

                # Get a function to execute in a subprocess to set the user and group IDs
                # to those of the given user.
                pre_exec = make_switch_user_closure(user)

            except TypeError:
                raise falcon.HTTPInternalServerError("Can't determine user.", "Unable to get the user name from the request.")

        try:
            # Don't run the subprocess with shell=True because that would introduce a security vulnerability.
            # A malicious user could POST a command like "show counters; rm -r ~".
            resp.body = check_output(('net {0}'.format(post_data['cmd'])).split(), stderr=STDOUT, preexec_fn=pre_exec)

#        except FileNotFoundError:  # TODO - Use this for Python3
        except OSError:  # TODO - This is for Python2.  Delete it for Python3, but be sure to catch PermissionError too.
            raise falcon.HTTPInternalServerError("Can't run NCLU.", 'Is NCLU installed on the host?')

        except CalledProcessError as e:
            # NCLU exited non-zero, which is okay.  This is the code path for an invalid 'net' command.
            resp.body = e.output

        # TODO - After the NCLU modularization work is complete, all commands will return json.
        if 'json' not in post_data['cmd'].lower():
            resp.content_type = 'text/plain'
        # Else, default to application/json.
