# Copyright (C) 2015, 2016, 2017 Cumulus Networks, Inc. All rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from gunicorn import util
from gunicorn.app.base import BaseApplication
from os.path import abspath, exists
from oslo.config import cfg
from restapi import settings
from restapi.utils.config import setup_conf
from restapi.utils.logs import get_logger, setup_logging
import six


_LOG = get_logger()


class _RestApiApplication(BaseApplication):
    """
    The python-cumulus-restapi package implements a WSGI application using
    the Falcon web framework.  The package has a runtime dependency on Gunicorn,
    a WSGI compatible server.  This class works with a Systemd service file
    to start Gunicorn with the appropriate settings.
    """

    def __init__(self):
        """
        Load the base configuration, and assign some core attributes.
        """
        setup_conf()
        setup_logging()
        super(_RestApiApplication, self).__init__()

    @staticmethod
    def get_config_from_filename(filename):
        gunicorn_cfg = {
            '__builtins__': __builtins__,
            '__file__': filename,
            '__name__': '__config__',
            '__doc__': None,
            '__package__': None
        }
        try:
            with open(filename) as f:
                try:
                    code = compile(f.read(), filename, 'exec')
                    six.exec_(code, gunicorn_cfg, gunicorn_cfg)
                except:
                    _LOG.error('Failed to read config file: %s', filename)
                    raise
        except OSError:
            _LOG.error('Failed to open config file: %s', filename)
        return gunicorn_cfg

    def load_config_from_filename(self, filename):
        gunicorn_cfg = self.get_config_from_filename(filename)
        for k, v in list(gunicorn_cfg.items()):
            # Ignore unknown names
            if k not in self.cfg.settings:
                continue
            try:
                self.cfg.set(k.lower(), v)
            except:
                _LOG.error('Invalid value for %s: %s' % (k, v))
                raise

    def load_config(self):
        self.init()
        if exists(cfg.CONF.server_config):
            self.load_config_from_filename(cfg.CONF.server_config)

    def load(self):
        return util.import_app('wsgi:api')

    def init(self, *args):
        self.cfg.set('bind', cfg.CONF.bind)
        self.cfg.set('pidfile', (
            self.cfg.pidfile or cfg.CONF.pidfile
        ))
        self.cfg.set('daemon', cfg.CONF.daemon)
        if settings.DEVELOPMENT:
            self.cfg.set('reload', True)
            self.cfg.set('timeout', 120)

    def run(self):
        try:
            self.load()
        except:
            _LOG.error('Error while loading the application')
            raise

        if self.cfg.daemon:
            util.daemonize(self.cfg.enable_stdio_inheritance)

        # set python paths
        if self.cfg.pythonpath and self.cfg.pythonpath is not None:
            paths = self.cfg.pythonpath.split(',')
            for path in paths:
                pythonpath = abspath(path)
                if pythonpath not in path:
                    path.insert(0, pythonpath)
        super(_RestApiApplication, self).run()


def main():
    app = _RestApiApplication()
    return app.run()
