# Copyright (C) 2015, 2016, 2017 Cumulus Networks, Inc. All rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from functools import wraps
from importlib import import_module
import collections
import re

_callable_cache = {}


def _memoize(cache, num_args):
    """Wrap a function so that results for any argument tuple are stored in
    'cache'. Note that the args to the function must be usable as dictionary
    keys.

    Only the first num_args are considered when creating the key.
    """
    def inner(func):
        @wraps(func)
        def wrapper(*args):
            mem_args = args[:num_args]
            if mem_args in cache:
                return cache[mem_args]
            result = func(*args)
            cache[mem_args] = result
            return result
        return wrapper
    return inner


@_memoize(_callable_cache, 1)
def get_callable(lookup_view, can_fail=False):
    """Convert a string version of a function name to the callable object.

    If the lookup_view is not an import path, it is assumed to be a URL pattern
    label and the original string is returned.

    If can_fail is True, lookup_view might be a URL pattern label, so errors
    during the import fail and the string is returned.
    """
    if not isinstance(lookup_view, collections.Callable):
        mod_name, func_name = get_mod_func(lookup_view)
        if func_name == '':
            return lookup_view

        try:
            mod = import_module(mod_name)
        except ImportError:
            raise
        else:
            try:
                lookup_view = getattr(mod, func_name)
                if not isinstance(lookup_view, collections.Callable):
                    raise RuntimeError(
                        "Could not import %s.%s. View is not callable." %
                        (mod_name, func_name))
            except AttributeError:
                if not can_fail:
                    raise RuntimeError(
                        "Could not import %s. View does not exist in "
                        "module %s." % (lookup_view, mod_name))
    return lookup_view


def get_mod_func(callback):
    # Converts 'django.views.news.stories.story_detail' to
    # ['django.views.news.stories', 'story_detail']
    try:
        dot = callback.rindex('.')
    except ValueError:
        return callback, ''
    return callback[:dot], callback[dot+1:]


def sanitize(input_str):
    return re.sub(r'[\t|\n]+', ' ', input_str)
