# Copyright (C) 2015, 2016, 2017 Cumulus Networks, Inc. All rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from importlib import import_module
from restapi.utils.misc import get_callable
import functools
import re
import six


class URLPattern(object):

    def __init__(self, uri, callback, name):
        self.uri = uri
        self.callback = callback
        self.name = name

    def add_prefix(self, prefix=None):
        if prefix:
            self.callback = prefix + '.' + self.callback

    def __repr__(self):
        return '<%s %s %s>' % (self.__class__.__name__,
                               self.name,
                               self.uri)


class URLResolver(object):

    def __init__(self, prefix, urlconf_name, app_name=None, namespace=None):
        self._prefix = prefix
        self.urlconf_name = urlconf_name
        if not isinstance(urlconf_name, six.string_types):
            self._urlconf_module = self.urlconf_name
        self.app_name = app_name
        self.namespace = namespace

    @property
    def prefix(self):
        return '%s/' % self._prefix if self._prefix else ''

    @property
    def urlconf_module(self):
        try:
            return self._urlconf_module
        except AttributeError:
            self._urlconf_module = import_module(self.urlconf_name)
            return self._urlconf_module

    @property
    def url_patterns(self):
        patterns = getattr(self.urlconf_module, "urlpatterns",
                           self.urlconf_module)
        try:
            iter(patterns)
        except TypeError:
            raise RuntimeError("The included urlconf %s doesn't have any "
                               "patterns in it" % self.urlconf_name)
        return patterns

    @property
    def version(self):
        version_str = getattr(self.urlconf_module, "version", None)
        return '%s/' % version_str if version_str else ''


def include(arg, namespace=None, app_name=None):
    if isinstance(arg, tuple):
        # callable returning a namespace hint
        if namespace:
            raise RuntimeError('Cannot override the namespace for a dynamic '
                               'module that provides a namespace')
        urlconf_module, app_name, namespace = arg
    else:
        # No namespace hint - use manually provided namespace
        urlconf_module = arg
    if isinstance(urlconf_module, six.string_types):
        urlconf_module = import_module(urlconf_module)
    patterns = getattr(urlconf_module, 'urlpatterns', urlconf_module)
    # Make sure we can iterate through the patterns (without this, some
    # testcases will break).
    if isinstance(patterns, (list, tuple)):
        for url_pattern in patterns:
            # Test if the URLResolver is used within the include;
            # this should throw an error since this is not allowed!
            if isinstance(url_pattern, URLResolver):
                raise RuntimeError('Using i18n_patterns in an included '
                                   'URLconf is not allowed.')
    return urlconf_module, app_name, namespace


def patterns(prefix, *args):
    pattern_list = []
    for t in args:
        if isinstance(t, (list, tuple)):
            t = url(prefix=prefix, *t)
        elif isinstance(t, URLPattern):
            t.add_prefix(prefix)
        pattern_list.append(t)
    return pattern_list


def url(regex, view, name=None, prefix=''):
    if isinstance(view, (list, tuple)):
        # For include(...) processing.
        urlconf_module, app_name, namespace = view
        return URLResolver(regex, urlconf_module, app_name=app_name,
                           namespace=namespace)
    else:
        if isinstance(view, six.string_types):
            if not view:
                raise RuntimeError('Empty URL pattern view name not permitted '
                                   '(for pattern %r)' % regex)
            if prefix:
                view = prefix + '.' + view
        return URLPattern(regex, view, name)


def reverse_route(urlpatterns, name, *args):
    def substitute_args(uri, *args):
        return functools.reduce(lambda p, arg: re.sub(r'{\w+}', arg, p, 1),
                                args, uri)

    for urlpattern in urlpatterns:
        if isinstance(urlpattern, URLResolver):
            pattern = reverse_route(urlpattern.url_patterns, name, *args)
            if pattern:
                return substitute_args(urlpattern.prefix +
                                       urlpattern.version + pattern,
                                       *args)
        elif urlpattern.name == name:  # URLPattern
            return substitute_args(urlpattern.uri, *args)
    else:
        raise RuntimeError('Route for %s not found' % name)


def expand_routes(urlpatterns):
    for urlpattern in urlpatterns:
        if isinstance(urlpattern, URLResolver):
            for pattern in urlpattern.url_patterns:
                yield (urlpattern.prefix + urlpattern.version + pattern.uri,
                       get_callable(pattern.callback), pattern.name)
        else:
            yield (urlpattern.uri, get_callable(urlpattern.callback),
                   urlpattern.name)
